from __future__ import annotations

import re
from .version import __version__ as version

# Session info code for python
# ------------------------------------------------------------
SESSION_INFO_PYTHON = r"""
# Inserted by pipen_runinfo, please do not modify
import atexit as _atexit


def _session_info(show_path: bool, include_submodule: bool):
    try:
        from importlib import metadata as importlib_metadata
    except ImportError:
        import importlib_metadata

    import sys
    import warnings
    {%% if "://" in str(job.metadir) %%}
    from yunpath import AnyPath as _AnyPath
    {%% else %%}
    from pathlib import Path as _AnyPath
    {%% endif %%}

    runinfo_file = _AnyPath("{{job.metadir}}/job.runinfo.session")

    lines = ["# Generated by pipen_runinfo v%(version)s\n", "# Lang: python\n"]
    if show_path:
        lines.append("Name\t__version__\timportlib.metadata\tPath\n")
        lines.append(f"python\t{sys.version}\t-\t{sys.executable}\n")
    else:
        lines.append("Name\t__version__\timportlib.metadata\n")
        lines.append(f"python\t{sys.version}\t-\n")

    for name, module in sys.modules.copy().items():
        if not include_submodule and "." in name:
            continue

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            ver = getattr(
                module,
                "__version__",
                getattr(module, "version", "-"),
            )
        mdfile = getattr(module, "__file__", None)
        if mdfile is None or "site-packages" not in mdfile or not module.__package__:
            # Suppose it's a built-in module
            continue

        try:
            imver = importlib_metadata.version(module.__package__)
        except importlib_metadata.PackageNotFoundError:
            imver = "-"

        if show_path:
            lines.append(f"{name}\t{ver}\t{imver}\t{mdfile}\n")
        else:
            lines.append(f"{name}\t{ver}\t{imver}\n")

    with runinfo_file.open("w") as fout:
        fout.writelines(lines)


@_atexit.register
def _run_session_info():
    _session_info(%(show_path)s, %(include_submodule)s)


# End of injected by pipen_runinfo
# ------------------------------------------------------------
# Regular script starts
# ------------------------------------------------------------
"""

future_import_statement = re.compile(
    r"^from\s+__future__\s+import\s+annotations\s*$",
    re.MULTILINE,
)


def inject_session_code_python(
    script: str,
    show_path: bool,
    include_submodule: bool,
) -> str:
    """Inject the session info code into a python script.

    Args:
        script: The script to inject the session info code into.
        show_path: Whether to include the path of the modules in the session info.
        include_submodule: Whether to include submodules in the session info.

    Returns:
        The injected script.
    """
    code = SESSION_INFO_PYTHON % {
        "version": version,
        "show_path": show_path,
        "include_submodule": include_submodule,
    }
    parts = future_import_statement.split(script, 1)
    if len(parts) == 1:
        return f"{code}\n\n{script}"

    # Matched!
    # Check if the future import is at the top
    # The first parts must be empty or comments
    if not parts[0].strip() or all(
        line.lstrip().startswith("#") or not line.strip()
        for line in parts[0].splitlines()
    ):
        return f"from __future__ import annotations\n{code}\n{''.join(parts)}"

    # The future import is not at the top, it is probably in a string
    return f"{code}\n\n{script}"


# Session info code for R
# ------------------------------------------------------------
SESSION_INFO_R = r"""
.save_session_info <- function() {
    .runinfo.session.file <- "{{job.metadir}}/job.runinfo.session"
    if (grepl("://", .runinfo.session.file)) {
        .runinfo.session.file.orig <- .runinfo.session.file
        .runinfo.session.file <- tempfile()
    }

    writeLines(
        c(
            "# Generated by pipen_runinfo v%(version)s",
            "# Lang: R",
            capture.output(sessionInfo())
        ),
        .runinfo.session.file
    )

    if (exists(".runinfo.session.file.orig")) {
        system2("cloudsh", c("mv", .runinfo.session.file, .runinfo.session.file.orig))
    }
}
""" % {"version": version}


def inject_session_code_r(
    script: str,
    show_path: bool,
    include_submodule: bool,
) -> str:
    # indent = " " * 4
    injected = [f"# Injected by pipen_runinfo v{version}, please do not modify"]
    injected.extend(SESSION_INFO_R.splitlines())
    injected.append("tryCatch({")
    injected.append("# End of injected by pipen_runinfo, please do not modify")
    injected.append("# ------------------------------------------------------")
    injected.append("# Regular script starts")
    injected.append("# ------------------------------------------------------")
    injected.append("")
    # injected.extend((f"{indent}{line}" for line in script.splitlines()))
    injected.append(script)
    injected.append("")
    injected.append("# ------------------------------------------------------")
    injected.append("# Regular script ends")
    injected.append("# ------------------------------------------------------")
    injected.append("# Injected by pipen_runinfo, please do not modify")
    injected.append("}, finally = .save_session_info())")
    injected.append("# End of injected by pipen_runinfo, please do not modify")
    injected.append("")
    return "\n".join(injected)


# Session info code for bash
# ------------------------------------------------------------
SESSION_INFO_BASH = r"""
# Injected by pipen_runinfo v%(version)s, please do not modify
_session_info() {
    runinfo_file="{{job.metadir}}/job.runinfo.session"
    if [[ "$runinfo_file" == *"://"* ]]; then
        runinfo_file_orig="$runinfo_file"
        runinfo_file=$(mktemp)
    fi

    echo "# Generated by pipen_runinfo v%(version)s" > $runinfo_file
    # shellcheck disable=SC2129
    echo "# Lang: bash" >> $runinfo_file
    echo -e "SHELL\t$SHELL" >> $runinfo_file
    echo -e "BASH_VERSION\t$BASH_VERSION" >> $runinfo_file
    echo -e "BASH_ARGV0\t$BASH_ARGV0" >> $runinfo_file
    echo -e "BASH_SOURCE\t$BASH_SOURCE" >> $runinfo_file
    exe=$(readlink /proc/$$/exe)
    # shellcheck disable=SC2129
    echo -e "proc-exe\t$exe" >> $runinfo_file
    echo -e "proc-exe-version\t$($exe --version | head -1)" >> $runinfo_file

    if [[ -v runinfo_file_orig ]]; then
        cloudsh mv "$runinfo_file" "$runinfo_file_orig"
    fi
}

trap _session_info EXIT
# End of injected by pipen_runinfo
# ------------------------------------------------------------
# Regular script starts
# ------------------------------------------------------------
""" % {"version": version}


def inject_session_code_bash(
    script: str,
    show_path: bool,
    include_submodule: bool,
) -> str:
    return f"{SESSION_INFO_BASH}\n\n{script}"


# Session info code for fish
# ------------------------------------------------------------
SESSION_INFO_FISH = r"""
# Injected by pipen_runinfo v%(version)s, please do not modify
function _session_info
    set -l runinfo_file "{{job.metadir}}/job.runinfo.session"
    if string match -q "*://*" $runinfo_file
        set runinfo_file_orig $runinfo_file
        set runinfo_file (mktemp)
    end

    echo "# Generated by pipen_runinfo v%(version)s" > $runinfo_file
    echo "# Lang: fish" >> $runinfo_file
    echo -e "SHELL\t$SHELL" >> $runinfo_file
    echo -e "FISH_VERSION\t$FISH_VERSION" >> $runinfo_file
    set -l exe (readlink /proc/(echo %%self)/exe)
    echo -e "proc-exe\t$exe" >> $runinfo_file
    echo -e "proc-exe-verion\t$($exe --version)" 2>/dev/null >> $runinfo_file

    if set -q runinfo_file_orig
        cloudsh mv $runinfo_file $runinfo_file_orig
    end
end

trap _session_info EXIT
# End of injected by pipen_runinfo, please do not modify
# ------------------------------------------------------------
# Regular script starts
# ------------------------------------------------------------
""" % {"version": version}


def inject_session_code_fish(
    script: str,
    show_path: bool,
    include_submodule: bool,
) -> str:
    return f"{SESSION_INFO_FISH}\n\n{script}"


def get_inject_session_code_fun(lang: str) -> str | None:
    """Get the language support class."""
    if lang == "python":
        return inject_session_code_python
    if lang == "R":
        return inject_session_code_r
    if lang == "bash":
        return inject_session_code_bash
    if lang == "fish":
        return inject_session_code_fish

    return None



from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING, List
from pathlib import Path

from yunpath import CloudPath
from pipen import plugin

from .version import __version__
from .session_info import get_inject_session_code_fun

if TYPE_CHECKING:  # pragma: no cover
    from pipen import Proc, Pipen
    from pipen.job import Job


def _get_lang(langpath: str | List[str]):
    if isinstance(langpath, list):
        langpath = langpath[0]

    stem = Path(langpath).stem if langpath else "bash"
    # Might be python3, python3.7, python3.7m, etc.
    if stem.startswith("python"):
        return "python"

    if stem in ("Rscript", "R") or stem.startswith("Rscript") or stem.startswith("R-"):
        return "R"

    if stem.startswith("bash"):
        return "bash"

    if stem.startswith("fish"):
        return "fish"

    return stem


class PipenRuninfoPlugin:
    name = "runinfo"
    version = __version__
    priority = 998

    @plugin.impl
    async def on_init(pipen: Pipen):
        """Called when the pipeline is initialized."""
        # Whether to include module path in the runinfo (for python only)
        # Either pipeline-level option or process-level option
        pipen.config.plugin_opts.setdefault("runinfo_path", True)
        # Whether to include submodules in the runinfo (for python only)
        # Either pipeline-level option or process-level option
        pipen.config.plugin_opts.setdefault("runinfo_submod", False)
        # Specify the lang directly instead of inferring from the proc.lang
        # Process-level option
        pipen.config.plugin_opts.setdefault("runinfo_lang", None)

    @plugin.impl
    def on_proc_script_computed(proc: Proc):
        """Called when a process is initialized.

        Try to modify the script so that we can get the runinfo.
        """
        pipeline_plugin_opts = proc.pipeline.config.get("plugin_opts", None) or {}
        proc_plugin_opts = proc.plugin_opts or {}
        runinfo_path = proc_plugin_opts.get(
            "runinfo_path",
            pipeline_plugin_opts.get("runinfo_path", True),
        )
        runinfo_submod = proc_plugin_opts.get(
            "runinfo_submod",
            pipeline_plugin_opts.get("runinfo_submod", False),
        )
        runinfo_lang = proc_plugin_opts.get(
            "runinfo_lang",
            pipeline_plugin_opts.get("runinfo_lang", None),
        )
        if not runinfo_lang:
            langpath = proc.lang
            runinfo_lang = _get_lang(langpath)

        if proc.script is None:  # pragma: no cover
            return

        inject_session_code_fun = get_inject_session_code_fun(runinfo_lang)
        if inject_session_code_fun is None:  # pragma: no cover
            return

        proc.script = inject_session_code_fun(
            proc.script,
            show_path=runinfo_path,
            include_submodule=runinfo_submod,
        )

    @plugin.impl
    def on_jobcmd_init(job: Job) -> str:
        if isinstance(job.metadir.mounted, CloudPath):  # pragma: no cover
            return textwrap.dedent(
                f"""
                # plugin: runinfo
                runinfo_device_orig="{job.metadir.mounted}/job.runinfo.device"
                runinfo_device=$(mktemp)
                runinfo_time_orig="{job.metadir.mounted}/job.runinfo.time"
                runinfo_time=$(mktemp)
                """
            )

        else:
            return textwrap.dedent(
                f"""
                # plugin: runinfo
                runinfo_device="{job.metadir.mounted}/job.runinfo.device"
                runinfo_time="{job.metadir.mounted}/job.runinfo.time"
                """
            )

    @plugin.impl
    def on_jobcmd_prep(job: Job) -> str:
        return textwrap.dedent(
            r"""
            # plugin: runinfo
            if env time -V &>/dev/null; then
                cmd="env time \
                    -f '# Generated by pipen-runinfo v%(version)s\n\n\
Command: %%C\n\
Voluntary context switches: %%w\n\
Involuntary context switches: %%c\n\
Percentage of CPU this job got: %%P\n\
Major page faults: %%F\n\
Minor page faults: %%R\n\
Maximum resident set size (kB): %%M\n\
Elapsed real time (s): %%e\n\
System (kernel) time (s): %%S\n\
User time (s): %%U\n\
Exit status: %%x' \
                    -o $runinfo_time $cmd"
            else
                echo "GNU time is not available, job is not timed." > $runinfo_time
                echo "See: https://www.gnu.org/software/time/" >> $runinfo_time
            fi
            """ % {"version": __version__}
        )

    @plugin.impl
    def on_jobcmd_end(job: Job) -> str:
        return textwrap.dedent(
            """        # plugin: runinfo

            echo "# Generated by pipen-runinfo v%(version)s" > $runinfo_device
            # shellcheck disable=SC2129
            echo "" >> $runinfo_device
            echo "Scheduler" >> $runinfo_device
            echo "---------" >> $runinfo_device
            echo "%(scheduler)s" >> $runinfo_device
            echo "" >> $runinfo_device
            echo "Hostname" >> $runinfo_device
            echo "--------" >> $runinfo_device
            hostname >> $runinfo_device
            echo "" >> $runinfo_device
            echo "CPU" >> $runinfo_device
            echo "----" >> $runinfo_device
            lscpu >> $runinfo_device
            echo "" >> $runinfo_device
            echo "Memory" >> $runinfo_device
            echo "------" >> $runinfo_device
            free -h >> $runinfo_device
            echo "" >> $runinfo_device
            echo "Disk" >> $runinfo_device
            echo "----" >> $runinfo_device
            df -h >> $runinfo_device
            echo "" >> $runinfo_device
            echo "Network" >> $runinfo_device
            echo "-------" >> $runinfo_device
            if ifconfig --version &>/dev/null; then
                ifconfig >> $runinfo_device
            else
                if ip -V &>/dev/null; then
                    ip a >> $runinfo_device
                else
                    echo "Neither ifconfig nor ip is available." >> $runinfo_device
                fi
            fi
            # shellcheck disable=SC2129
            echo "" >> $runinfo_device
            echo "GPU" >> $runinfo_device
            echo "---" >> $runinfo_device
            if nvidia-smi --version &>/dev/null; then
                nvidia-smi >> $runinfo_device
            else
                echo "nvidia-smi is not available." >> $runinfo_device
            fi
            echo "" >> $runinfo_device

            if [[ -v runinfo_device_orig ]]; then
                cloudsh mv $runinfo_device $runinfo_device_orig
                cloudsh mv $runinfo_time $runinfo_time_orig
            fi
            """ % {"scheduler": job.proc.scheduler.name, "version": __version__}
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pipen_runinfo']

package_data = \
{'': ['*']}

install_requires = \
['pipen==0.17.*']

entry_points = \
{'pipen': ['runinfo = pipen_runinfo:PipenRuninfoPlugin']}

setup_kwargs = {
    'name': 'pipen-runinfo',
    'version': '0.9.2',
    'description': 'Generate running information for jobs in pipen pipelines',
    'long_description': '# pipen-runinfo\n\nGenerate running information for jobs in [pipen][1] pipelines.\n\nIncluding session info (packages and versions), time, and device info.\n\n## Install\n\n```bash\npip install -U pipen-runinfo\n```\n\n## Enable/Disable the plugin\n\nThe plugin is registered via entrypoints. It\'s by default enabled. To disable it: plugins=[..., "no:runinfo"], or uninstall this plugin.\n\n## Plugin options\n\n- `runinfo_lang`: The name of the language to run the job script for session info.\n    Default is `None`, which means it will be inferred from the `proc.lang`\n    This should be a process-level option, unless you only have one single\n    process in your pipeline.\n- `runinfo_path`: Whether to include paths for the modules in the session information.\n    Default is `True`.\n    This option could be either specified in the process-level or the pipeline-level.\n    Only works for `python`.\n- `runinfo_submod`: Whether to include submodules in the session information.\n    Default is `False`.\n    This option could be either specified in the process-level or the pipeline-level.\n    Only works for `python`.\n\n## Supported languages for session info\n\n`python`, `R`, `bash`, and `fish`.\n\n## Usage\n\nThe plugin will generate 3 files in the job directory of the pipeline.\n\n### `job.runinfo.session`\n\nThe session information of the job, including the interpreter, packages and their versions.\n\n#### Python\n\nGenerates a TSV file with the following columns:\n\n- `Name`: The name of the module, or python itself\n- `__version__`: The version fetched by `module.__version__` or `module.version`\n- `importlib.metadata`: The version fetched by `importlib.metadata.version(package)`\n- `Path`: The path of the module (only if `runinfo_path` is `True`)\n\n#### R\n\nGenerates a text file `sessionInfo()` output.\n\n#### Bash\n\nGenerates a TSV file with the following columns:\n\n- `SHELL`: The value of `$SHELL`\n- `BASH_VERSION`: The value of `$BASH_VERSION`\n- `BASH_ARGV0`: The value of `$BASH_ARGV0`\n- `BASH_SOURCE`: The value of `$BASH_SOURCE`\n- `proc-exe`: The real path of the executable from `/proc/<pid>/exe`\n- `proc-exe-version`: The version of the executable from `/proc/<pid>/exe --version`\n\n#### Fish\n\nGenerates a TSV file with the following columns:\n\n- `SHELL`: The value of `$SHELL`\n- `FISH_VERSION`: The value of `$FISH_VERSION`\n- `proc-exe`: The real path of the executable from `/proc/<pid>/exe`\n- `proc-exe-version`: The version of the executable from `/proc/<pid>/exe --version`\n\n### `job.runinfo.time`\n\nThe time spent on the job, and more, generated by `time -v` command.\n\n### `job.runinfo.device`\n\nThe device (cpu and memory) information of the job, generated by `lscpu`/`lsmem` command.\n\n\n[1]: https://github.com/pwwang/pipen\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
